# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base for HTML controls."""

import abc
import contextlib
import inspect
import sys
from typing import Annotated, Any, Dict, Iterator, List, Optional, Union

from pyglove.core import utils
from pyglove.core.symbolic import object as pg_object
from pyglove.core.views.html import base

Html = base.Html


try:
  _notebook = sys.modules['IPython'].display
except Exception:  # pylint: disable=broad-except
  _notebook = None


class HtmlControl(pg_object.Object):
  """Base class for HTML controls."""

  id: Annotated[
      Optional[str],
      'Optional ID of the root element.'
  ] = None

  css_classes: Annotated[
      List[str],
      'Optional CSS classes of the root element.'
  ] = []

  styles: Annotated[
      Dict[str, Any],
      'Ad-hoc styles (vs. class-based) of the root element.'
  ] = {}

  interactive: Annotated[
      bool,
      (
          'Whether the control is interactive. If False, optimizations will be '
          'applied to reduce the size of the HTML.'
      )
  ] = False

  def _on_bound(self):
    super()._on_bound()
    self._rendered = False
    self._css_styles = []
    self._dynamic_injected_css = set()
    self._scripts = []

  def add_style(self, *css: str) -> 'HtmlControl':
    """Adds CSS styles to the HTML."""
    self._css_styles.extend(css)
    return self

  def add_script(self, *scripts: str) -> 'HtmlControl':
    self._scripts.extend(scripts)
    return self

  def to_html(self, **kwargs) -> Html:
    """Returns the HTML representation of the control."""
    self._rendered = True
    self._dynamic_injected_css = set()
    html = self._to_html(**kwargs)
    return html.add_style(*self._css_styles).add_script(*self._scripts)

  @abc.abstractmethod
  def _to_html(self, **kwargs) -> Html:
    """Returns the HTML representation of the control."""

  @classmethod
  @contextlib.contextmanager
  def track_scripts(cls) -> Iterator[List[str]]:
    del cls
    all_tracked = utils.thread_local_get(_TLS_TRACKED_SCRIPTS, [])
    current = []
    all_tracked.append(current)
    utils.thread_local_set(_TLS_TRACKED_SCRIPTS, all_tracked)
    try:
      yield current
    finally:
      all_tracked.pop(-1)
      if not all_tracked:
        utils.thread_local_del(_TLS_TRACKED_SCRIPTS)

  def _sync_members(self, **fields) -> None:
    """Synchronizes displayed values to members."""
    self.rebind(fields, skip_notification=True, raise_on_no_change=False)

  def _run_javascript(self, code: str, debug: bool = False) -> None:
    """Runs the given JavaScript code."""
    if not self.interactive:
      raise ValueError(
          f'Non-interactive control {self} cannot be updated. '
          'Please set `interactive=True` in the constructor.'
      )
    if not self._rendered:
      return

    code = inspect.cleandoc(code)
    if debug:
      print('RUN JAVSCRIPT:\n', code)
    if _notebook is not None:
      _notebook.display(_notebook.Javascript(code))

    # Track script execution.
    all_tracked = utils.thread_local_get(_TLS_TRACKED_SCRIPTS, [])
    for tracked in all_tracked:
      tracked.append(code)

  def _add_css_rules(self, css: str) -> None:
    if not self._rendered or not css or css in self._dynamic_injected_css:
      return
    self._run_javascript(
        f"""
        const style = document.createElement('style');
        style.type = 'text/css';
        style.textContent = "{Html.escape(css, javascript_str=True)}";
        document.head.appendChild(style);
        """
    )
    self._dynamic_injected_css.add(css)

  def _apply_css_rules(self, html: Html) -> None:
    self._add_css_rules(html.styles.content)

  def _insert_adjacent_html(
      self,
      element_selector_js: str,
      html: Html,
      var_name: str = 'elem',
      position: str = 'beforeend'
  ):
    self._run_javascript(
        f"""
        {element_selector_js}
        {var_name}.insertAdjacentHTML(
            "{position}",
            "{Html.escape(html, javascript_str=True).to_str(content_only=True)}"
        );
        """,
    )
    self._apply_css_rules(html)

  def element_id(self, child: Optional[str] = None) -> Optional[str]:
    """Returns the element id of this control or a child."""
    if self.id is not None:
      return self.id
    elif not self.interactive:
      return None
    elif child is None:
      return f'control-{id(self)}'
    else:
      return f'control-{id(self)}-{child}'

  def _update_content(
      self,
      content: Union[str, Html],
      child: Optional[str] = None,
  ) -> None:
    """Updates the content of the control."""
    if isinstance(content, str):
      return self._update_text(content, child)
    else:
      return self._update_inner_html(content, child)

  def _update_text(
      self,
      content: str,
      child: Optional[str] = None
  ) -> str:
    """Updates the content of the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.textContent = "{Html.escape(content, javascript_str=True)}";
        """
    )
    return content

  def _update_inner_html(
      self,
      html: base.Html,
      child: Optional[str] = None,
  ) -> base.Html:
    """Updates the inner HTML of the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.innerHTML = "{Html.escape(html, javascript_str=True).to_str(content_only=True)}";
        """
    )
    self._add_css_rules(html.styles.content)
    return html

  def _update_style(
      self,
      styles: Dict[str, Any],
      *,
      child: Optional[str] = None,
      updates_only: bool = False,
  ) -> Dict[str, Any]:
    """Updates the style of the control."""
    updated_styles = {} if updates_only else dict(self.styles or {})
    updated_styles.update(styles)
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.style = "{Html.style_str(updated_styles)}";
        """
    )
    return updated_styles

  def _update_property(
      self,
      name: str,
      value: str,
      child: Optional[str] = None,
  ) -> str:
    """Updates a property of the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.{name} = "{value}";
        """
    )
    return value

  def _add_css_class(
      self,
      css_class: str,
      child: Optional[str] = None,
  ) -> str:
    """Adds a CSS class to the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.classList.add("{css_class}");
        """
    )
    return css_class

  def _remove_css_class(
      self,
      css_class: str,
      child: Optional[str] = None,
  ) -> str:
    """Removes a CSS class from the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.classList.remove("{css_class}");
        """
    )
    return css_class


_TLS_TRACKED_SCRIPTS = '__tracked_scripts__'



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.hyper.Float."""

import unittest

from pyglove.core import geno
from pyglove.core.hyper.categorical import oneof
from pyglove.core.hyper.dynamic_evaluation import trace as pg_trace
from pyglove.core.hyper.iter import iterate as pg_iterate
from pyglove.core.hyper.iter import random_sample as pg_random_sample


class IterateTest(unittest.TestCase):
  """Tests for pg.iter."""

  def test_iter_with_default_algorithm(self):
    v = oneof(range(100))
    examples = list(pg_iterate(v))
    self.assertEqual(examples, list(range(100)))

    examples = list(pg_iterate(v, 10))
    self.assertEqual(examples, list(range(10)))

  def test_iter_with_custom_algorithm(self):

    class ConstantAlgorithm(geno.DNAGenerator):

      def _on_bound(self):
        self._rewards = []

      def _propose(self):
        if len(self._rewards) == 100:
          raise StopIteration()
        return geno.DNA(0)

      def _feedback(self, dna, reward):
        self._rewards.append(reward)

      @property
      def rewards(self):
        return self._rewards

    algo = ConstantAlgorithm()
    examples = []
    for i, (x, feedback) in enumerate(pg_iterate(oneof([1, 3]), 5, algo)):
      examples.append(x)
      feedback(float(i))
      self.assertEqual(feedback.dna, geno.DNA(0))
    self.assertEqual(len(examples), 5)
    self.assertEqual(examples, [1] * 5)
    self.assertEqual(algo.rewards, [float(i) for i in range(5)])

    for x, feedback in pg_iterate(oneof([1, 3]), algorithm=algo):
      examples.append(x)
      feedback(0.)
    self.assertEqual(len(examples), 100)

  def test_iter_with_dynamic_evaluation(self):
    def foo():
      return oneof([1, 3])
    examples = []
    for x in pg_iterate(pg_trace(foo)):
      with x():
        examples.append(foo())
    self.assertEqual(examples, [1, 3])

  def test_iter_with_continuation(self):

    class ConstantAlgorithm3(geno.DNAGenerator):

      def setup(self, dna_spec):
        super().setup(dna_spec)
        self.num_trials = 0

      def propose(self):
        self.num_trials += 1
        return geno.DNA(0)

    algo = ConstantAlgorithm3()
    for unused_x in pg_iterate(oneof([1, 3]), 10, algo):
      pass
    for unused_x in pg_iterate(oneof([1, 3]), 10, algo):
      pass
    self.assertEqual(algo.num_trials, 20)

  def test_iter_with_forced_feedback(self):

    class ConstantAlgorithm2(geno.DNAGenerator):

      def propose(self):
        return geno.DNA(0)

    algo = ConstantAlgorithm2()
    examples = []
    for x, feedback in pg_iterate(
        oneof([1, 3]), 10, algorithm=algo, force_feedback=True):
      examples.append(x)
      # No op.
      feedback(0.)
    self.assertEqual(len(examples), 10)

  def test_bad_iter(self):
    with self.assertRaisesRegex(
        ValueError, '\'hyper_value\' is a constant value'):
      next(pg_iterate('foo'))

    algo = geno.Random()
    next(pg_iterate(oneof([1, 2]), 1, algo))
    with self.assertRaisesRegex(
        ValueError, '.* has been set up with a different DNASpec'):
      next(pg_iterate(oneof([2, 3]), 10, algo))


class RandomSampleTest(unittest.TestCase):
  """Tests for pg.random_sample."""

  def test_random_sample(self):
    self.assertEqual(
        list(pg_random_sample(oneof([0, 1]), 3, seed=123)), [0, 1, 0])


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for timing."""

import collections
import dataclasses
import time
from typing import Any, Dict, List, Optional

from pyglove.core.utils import error_utils
from pyglove.core.utils import json_conversion
from pyglove.core.utils import thread_local


class TimeIt:
  """Context manager for timing the execution of a code block."""

  @dataclasses.dataclass(frozen=True)
  class Status(json_conversion.JSONConvertible):
    """Status of a single `pg.timeit`."""
    name: str
    elapse: float = 0.0
    has_ended: bool = True
    error: Optional[error_utils.ErrorInfo] = None

    @property
    def has_started(self) -> bool:
      """Returns whether the context has started."""
      return self.elapse > 0

    @property
    def has_error(self) -> bool:
      """Returns whether the context has error."""
      return self.error is not None

    def to_json(self, **kwargs) -> Dict[str, Any]:
      return self.to_json_dict(
          fields=dict(
              name=(self.name, None),
              elapse=(self.elapse, 0.0),
              has_ended=(self.has_ended, True),
              error=(self.error, None),
          ),
          exclude_default=True,
          **kwargs,
      )

    def merge(self, other: 'TimeIt.Status') -> 'TimeIt.Status':
      """Merges the status of two `pg.timeit`."""
      assert other.name == self.name, (self.name, other.name)
      return TimeIt.Status(
          name=self.name,
          elapse=self.elapse + other.elapse,
          has_ended=self.has_ended and other.has_ended,
          error=self.error or other.error,
      )

  @dataclasses.dataclass
  class StatusSummary(json_conversion.JSONConvertible):
    """Aggregated summary for repeated calls for `pg.timeit`."""

    @dataclasses.dataclass
    class Entry(json_conversion.JSONConvertible):
      """Aggregated status from the `pg.timeit` calls of the same name."""

      num_started: int = 0
      num_ended: int = 0
      num_failed: int = 0
      avg_duration: float = 0.0
      error_tags: Dict[str, int] = dataclasses.field(
          default_factory=lambda: collections.defaultdict(int)
      )

      def update(self, status: 'TimeIt.Status'):
        self.avg_duration = (
            (self.avg_duration * self.num_started + status.elapse)
            / (self.num_started + 1)
        )
        self.num_started += 1
        if status.has_ended:
          self.num_ended += 1
        if status.has_error:
          self.num_failed += 1
          assert status.error is not None
          self.error_tags[status.error.tag] += 1

      def to_json(self, **kwargs) -> Dict[str, Any]:
        return self.to_json_dict(
            fields=dict(
                num_started=(self.num_started, 0),
                num_ended=(self.num_ended, 0),
                num_failed=(self.num_failed, 0),
                avg_duration=(self.avg_duration, 0.0),
                error_tags=(self.error_tags, {}),
            ),
            exclude_default=True,
            **kwargs,
        )

    breakdown: dict[str, 'TimeIt.StatusSummary.Entry'] = (
        dataclasses.field(default_factory=dict)
    )

    def __bool__(self) -> bool:
      """Returns True if the summary is non-empty."""
      return bool(self.breakdown)

    def aggregate(self, timeit_status: Dict[str, 'TimeIt.Status']):
      for k, v in timeit_status.items():
        if k not in self.breakdown:
          self.breakdown[k] = TimeIt.StatusSummary.Entry()
        self.breakdown[k].update(v)

    def to_json(self, **kwargs) -> Dict[str, Any]:
      return self.to_json_dict(
          fields=dict(
              breakdown=(self.breakdown, {}),
          ),
          exclude_default=True,
          **kwargs,
      )

  def __init__(self, name: str = ''):
    self._name: str = name
    self._start_time: Optional[float] = None
    self._end_time: Optional[float] = None
    self._child_contexts: List[TimeIt] = []
    self._error: Optional[error_utils.ErrorInfo] = None
    self._parent: Optional[TimeIt] = None

  @property
  def name(self) -> str:
    """Returns the name of the context."""
    return self._name

  @property
  def children(self) -> List['TimeIt']:
    """Returns child contexts."""
    return self._child_contexts

  def add(self, context: 'TimeIt'):
    """Adds a child context."""
    self._child_contexts.append(context)

  def start(self):
    """Starts timing."""
    self._start_time = time.time()

  def end(self, error: Optional[BaseException] = None) -> bool:
    """Ends timing."""
    if not self.has_ended:
      self._end_time = time.time()
      self._error = (
          None if error is None else error_utils.ErrorInfo.from_exception(error)
      )
      return True
    return False

  @property
  def has_started(self) -> bool:
    """Returns whether the context has started."""
    return self._start_time is not None

  @property
  def has_ended(self) -> bool:
    """Returns whether the context has ended."""
    return self._end_time is not None

  @property
  def start_time(self) -> Optional[float]:
    """Returns start time."""
    return self._start_time

  @property
  def end_time(self) -> Optional[float]:
    """Returns end time."""
    return self._end_time

  @property
  def error(self) -> Optional[error_utils.ErrorInfo]:
    """Returns error."""
    return self._error

  @property
  def has_error(self) -> bool:
    """Returns whether the context has error."""
    return self._error is not None

  @property
  def elapse(self) -> float:
    """Returns the elapse since start until end."""
    if self._start_time is None:
      return 0
    if self._end_time is None:
      return time.time() - self._start_time
    return self._end_time - self._start_time  # pytype: disable=unsupported-operands

  def status(self) -> Dict[str, Status]:
    """Gets the status of all `timeit` under this context."""
    result = {
        self.name: TimeIt.Status(
            name=self.name, elapse=self.elapse,
            has_ended=self.has_ended, error=self._error,
        )
    }
    for child in self._child_contexts:
      child_result = child.status()
      for k, v in child_result.items():
        key = f'{self.name}.{k}' if self.name else k
        if key in result:
          result[key] = result[key].merge(v)
        else:
          result[key] = v
    return result

  def __enter__(self):
    parent = thread_local.thread_local_get('__timing_context__', None)
    if parent is not None:
      parent.add(self)
      self._parent = parent
    thread_local.thread_local_set('__timing_context__', self)
    self.start()
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    del exc_type, traceback
    self.end(exc_value)
    if self._parent is None:
      thread_local.thread_local_del('__timing_context__')
    else:
      thread_local.thread_local_set('__timing_context__', self._parent)


def timeit(name: str = '') -> TimeIt:
  """Context manager to time a block of code."""
  return TimeIt(name)



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tab control."""

from typing import Annotated, List, Literal, Optional, Union

from pyglove.core.symbolic import flags as pg_flags
from pyglove.core.symbolic import object as pg_object

# pylint: disable=g-importing-member
from pyglove.core.views.html.base import Html
from pyglove.core.views.html.base import HtmlConvertible
from pyglove.core.views.html.controls.base import HtmlControl
from pyglove.core.views.html.controls.label import Label
# pylint: enable=g-importing-member


class Tab(pg_object.Object):
  """A tab."""

  label: Annotated[
      Label,
      'The label of the tab.'
  ]

  content: Annotated[
      Union[Html, HtmlConvertible],
      'The content of the tab.'
  ]

  css_classes: Annotated[
      List[str],
      'The CSS classes of the tab.'
  ] = []

  name: Annotated[
      Optional[str],
      'An optional name that can be used to identify a tab under a tab control'
  ] = None


@pg_object.use_init_args(
    ['tabs', 'selected', 'tab_position', 'id', 'css_classes', 'styles']
)
class TabControl(HtmlControl):
  """A tab control."""

  tabs: Annotated[
      List[Tab],
      'The tabs of the tab control.'
  ]

  selected: Annotated[
      int,
      'The index of the selected tab.'
  ] = 0

  tab_position: Annotated[
      Literal['top', 'left'],
      'The direction of the tab control.'
  ] = 'top'

  interactive = True

  def append(self, tab: Tab) -> None:
    with pg_flags.notify_on_change(False):
      self.tabs.append(tab)

    self._insert_adjacent_html(
        f"""
        const elem = document.getElementById('{self.element_id()}-button-group');
        """,
        self._tab_button(tab, len(self.tabs) - 1),
        position='beforeend',
    )
    self._insert_adjacent_html(
        f"""
        const elem = document.getElementById('{self.element_id()}-content-group');
        """,
        self._tab_content(tab, len(self.tabs) - 1),
        position='beforeend',
    )

  def insert(self, index_or_name: Union[int, str], tab: Tab) -> None:
    """Inserts a tab before a tab identified by index or name."""
    index = self.indexof(index_or_name)
    if index == -1:
      raise ValueError(f'Tab not found: {index_or_name!r}')
    with pg_flags.notify_on_change(False):
      self.tabs.insert(index, tab)

    self._insert_adjacent_html(
        f"""
        const elem = document.querySelectorAll('#{self.element_id()}-button-group > .tab-button')[{index}];
        """,
        self._tab_button(tab, len(self.tabs) - 1),
        position='beforebegin',
    )
    self._insert_adjacent_html(
        f"""
        const elem = document.querySelectorAll('#{self.element_id()}-content-group > .tab-content')[{index}];
        """,
        self._tab_content(tab, len(self.tabs) - 1),
        position='beforebegin',
    )

  def indexof(self, index_or_name: Union[int, str]) -> int:
    if isinstance(index_or_name, int):
      index = index_or_name
      if index >= len(self.tabs):
        return len(self.tabs) - 1
      elif index < -len(self.tabs):
        return -1
      elif index < 0:
        index = index + len(self.tabs)
      return index
    else:
      name = index_or_name
      assert isinstance(name, str), name
      for i, tab in enumerate(self.tabs):
        if tab.name == name:
          return i
      return -1

  def extend(self, tabs: List[Tab]) -> None:
    for tab in tabs:
      self.append(tab)

  def select(
      self,
      index_or_name: Union[int, str, List[str]]) -> Union[int, str]:
    """Selects a tab identified by an index or name.

    Args:
      index_or_name: The index or name of the tab to select. If a list of names
        is provided, the first name in the list that is found will be selected.

    Returns:
      The index (if the index was provided) or name of the selected tab.
    """
    selected_name = index_or_name if isinstance(index_or_name, str) else None
    index = -1
    if isinstance(index_or_name, list):
      for name in index_or_name:
        index = self.indexof(name)
        if index != -1:
          selected_name = name
          break
    else:
      index = self.indexof(index_or_name)
    if index == -1:
      raise ValueError(f'Tab not found: {index_or_name!r}')
    self._sync_members(selected=index)
    self._run_javascript(
        f"""
        const tabButtons = document.querySelectorAll('#{self.element_id()}-button-group > .tab-button');
        tabButtons[{index}].click();
        """
    )
    return selected_name or index

  def _to_html(self, **kwargs):
    return Html.element(
        'table',
        [
            '<tr><td>',
            Html.element(
                'div',
                [self._tab_button(tab, i) for i, tab in enumerate(self.tabs)],
                css_classes=[
                    'tab-button-group',
                    self.tab_position
                ] + self.css_classes,
                id=self.element_id('button-group'),
            ),
            ('</td><td>' if self.tab_position == 'left'
             else '</td></tr><tr><td>'),
            Html.element(
                'div',
                [self._tab_content(tab, i) for i, tab in enumerate(self.tabs)],
                css_classes=[
                    'tab-content-group',
                    self.tab_position
                ] + self.css_classes,
                id=self.element_id('content-group'),
            ),
            '</td></tr>'
        ],
        css_classes=['tab-control'],
        styles=self.styles,
    ).add_script(
        """
        function openTab(event, controlId, tabId) {
          const tabButtons = document.querySelectorAll('#' + controlId + '-button-group > .tab-button');
          for (let i = 0; i < tabButtons.length; i++) {
            tabButtons[i].classList.remove('selected');
          }
          const tabContents = document.querySelectorAll('#' + controlId + '-content-group > .tab-content');
          for (let i = 0; i < tabContents.length; i++) {
            tabContents[i].classList.remove('selected')
          }
          const tabButton = event.currentTarget;
          tabButton.classList.add('selected');
          document.getElementById(tabId).classList.add('selected');
        }
        """
    ).add_style(
        """
        .tab-control {
          border-spacing: 0px;
          border-collapse: collapse;
          margin-top: 10px;
        }
        .tab-control td {
          padding: 0px;
          margin: 0px;
          vertical-align: top;
        }
        .top.tab-button-group {
          border-left: 1px solid #DDD;
          border-top: 1px solid #DDD;
          border-right: 1px solid #DDD;
          border-radius: 5px 5px 0px 0px;
          padding: 0px 5px 0px 0px;
          margin-bottom: -2px;
        }
        .tab-button {
          background-color: transparent;
          border-radius: 5px;
          border: 0px;
          font-weight: bold;
          color: gray;
          outline: none;
          cursor: pointer;
          transition: 0.3s;
        }
        .tab-button:hover {
          background-color: #fff1dd;
        }
        .tab-button.selected {
          background-color: #f0ecf9;
          color: black;
        }
        .top.tab-content-group {
          border-left: 1px solid #DDD;
          border-right: 1px solid #DDD;
          border-bottom: 1px solid #DDD;
          border-radius: 0px 0px 5px 5px;
          margin: 0px;
          padding: 5px;
          height: 100%;
        }
        .top > .tab-button {
          margin: 5px 0px 5px 5px;
        }
        .tab-content {
          display: none;
        }
        .tab-content.selected {
          display: block;
        }
        .left.tab-button-group {
          display: inline-flex;
          flex-direction: column;
          border: 1px solid #DDD;
          border-radius: 5px;
          margin-right: 5px;
          padding: 0px 0px 5px 0px;
        }
        .left.tab-content-group {
          border: 0px
          margin: 0px;
          padding: 0px;
          height: 100%;
        }
        .left > .tab-button {
          text-align: left;
          margin: 5px 5px 0px 5px;
        }
        """
    )

  def _tab_button(self, tab: Tab, i: int) -> Html:
    return Html.element(
        'button',
        [
            tab.label
        ],
        css_classes=[
            'tab-button',
            'selected' if i == self.selected else None
        ] + tab.css_classes,
        onclick=(
            f"""openTab(event, '{self.element_id()}', '{self.element_id(str(i))}')"""
        )
    )

  def _tab_content(self, tab: Tab, i: int) -> Html:
    return Html.element(
        'div',
        [
            tab.content
        ],
        css_classes=[
            'tab-content',
            'selected' if i == self.selected else None
        ] + tab.css_classes,
        id=self.element_id(str(i))
    )



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import sys
import typing
from typing import List, Literal, Union
import unittest

from pyglove.core import symbolic as pg
from pyglove.core.typing import key_specs as ks
from pyglove.core.typing import value_specs as vs


class AnnotationFutureConversionTest(unittest.TestCase):

  # Class with forward declaration must not be defined in functions.
  class A(pg.Object):
    a: typing.Optional[AnnotationFutureConversionTest.A]
    b: List[AnnotationFutureConversionTest.A]

  def assert_value_spec(self, cls, field_name, expected_value_spec):
    self.assertEqual(cls.__schema__[field_name].value, expected_value_spec)

  def test_basics(self):

    class Foo(pg.Object):
      a: int
      b: float
      c: bool
      d: str
      e: typing.Any
      f: typing.Dict[str, typing.Any]
      g: typing.List[str]
      h: typing.Tuple[int, int]
      i: typing.Callable[[int, int], None]

    self.assert_value_spec(Foo, 'a', vs.Int())
    self.assert_value_spec(Foo, 'b', vs.Float())
    self.assert_value_spec(Foo, 'c', vs.Bool())
    self.assert_value_spec(Foo, 'd', vs.Str())
    self.assert_value_spec(Foo, 'e', vs.Any(annotation=typing.Any))
    self.assert_value_spec(
        Foo, 'f', vs.Dict([(ks.StrKey(), vs.Any(annotation=typing.Any))])
    )
    self.assert_value_spec(Foo, 'g', vs.List(vs.Str()))
    self.assert_value_spec(Foo, 'h', vs.Tuple([vs.Int(), vs.Int()]))
    self.assert_value_spec(
        Foo, 'i',
        vs.Callable([vs.Int(), vs.Int()], returns=vs.Object(type(None)))
    )

  def test_list(self):
    if sys.version_info >= (3, 10):

      class Bar(pg.Object):
        x: list[int | None]

      self.assert_value_spec(Bar, 'x', vs.List(vs.Int().noneable()))

  def test_var_length_tuple(self):

    class Foo(pg.Object):
      x: typing.Tuple[int, ...]

    self.assert_value_spec(Foo, 'x', vs.Tuple(vs.Int()))

    if sys.version_info >= (3, 10):

      class Bar(pg.Object):
        x: tuple[int, ...]

      self.assert_value_spec(Bar, 'x', vs.Tuple(vs.Int()))

  def test_optional(self):

    class Foo(pg.Object):
      x: typing.Optional[int]

    self.assert_value_spec(Foo, 'x', vs.Int().noneable())

    if sys.version_info >= (3, 10):
      class Bar(pg.Object):
        x: int | None

      self.assert_value_spec(Bar, 'x', vs.Int().noneable())

  def test_union(self):

    class Foo(pg.Object):
      x: Union[int, typing.Union[str, bool], None]

    self.assert_value_spec(
        Foo, 'x', vs.Union([vs.Int(), vs.Str(), vs.Bool()]).noneable()
    )

    if sys.version_info >= (3, 10):

      class Bar(pg.Object):
        x: int | str | bool

      self.assert_value_spec(
          Bar, 'x', vs.Union([vs.Int(), vs.Str(), vs.Bool()])
      )

  def test_literal(self):

    class Foo(pg.Object):
      x: Literal[1, True, 'abc']

    self.assert_value_spec(
        Foo, 'x', vs.Enum(vs.MISSING_VALUE, [1, True, 'abc'])
    )

  def test_self_referencial(self):
    self.assert_value_spec(
        self.A, 'a', vs.Object(self.A).noneable()
    )
    self.assert_value_spec(
        self.A, 'b', vs.List(vs.Object(self.A))
    )

if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from typing import Any
import unittest
import weakref

from pyglove.core import utils as pg_utils
from pyglove.core.symbolic import contextual_object

contextual_override = pg_utils.contextual_override
ContextualOverride = pg_utils.ContextualOverride
get_contextual_override = pg_utils.get_contextual_override
contextual_value = pg_utils.contextual_value
all_contextual_values = pg_utils.all_contextual_values

ContextualObject = contextual_object.ContextualObject
contextual_attribute = contextual_object.contextual_attribute


class ContextualObjectTest(unittest.TestCase):
  """Tests for ContextualObject."""

  def test_override(self):
    class A(ContextualObject):
      x: int

    a = A(x=1)
    with a.override(x=2, y=1):
      self.assertEqual(a.x, 2)

      # `y`` is not an attribute of `A`.
      with self.assertRaises(AttributeError):
        _ = a.y

  def test_context(self):
    class A(ContextualObject):
      x: int
      y: int = contextual_attribute()
      z: int = contextual_attribute(default=-1)

    with self.assertRaisesRegex(TypeError, '.* missing 1 required argument'):
      _ = A()

    a = A(x=1)
    with self.assertRaisesRegex(
        AttributeError, 'p'
    ):
      _ = a.p

    with self.assertRaisesRegex(
        AttributeError, '.* is not found under its context'
    ):
      _ = a.y

    with contextual_override(y=1):
      self.assertEqual(a.y, 1)

    with self.assertRaisesRegex(
        AttributeError, '.* is not found under its context'
    ):
      _ = a.y

    # Use contextual default if it's not provided.
    self.assertEqual(a.z, -1)

    a1 = A(x=1, y=2)
    self.assertEqual(a1.x, 1)
    self.assertEqual(a1.y, 2)
    self.assertEqual(a1.z, -1)

    with contextual_override(x=3, y=3, z=3) as parent_override:
      self.assertEqual(
          parent_override,
          dict(
              x=ContextualOverride(3, cascade=False, override_attrs=False),
              y=ContextualOverride(3, cascade=False, override_attrs=False),
              z=ContextualOverride(3, cascade=False, override_attrs=False),
          ),
      )
      self.assertEqual(
          get_contextual_override('y'),
          ContextualOverride(3, cascade=False, override_attrs=False),
      )
      self.assertEqual(contextual_value('x'), 3)
      self.assertIsNone(contextual_value('f', None))
      with self.assertRaisesRegex(KeyError, '.* does not exist'):
        contextual_value('f')

      self.assertEqual(all_contextual_values(), dict(x=3, y=3, z=3))

      # Member attributes take precedence over `contextual_override`.
      self.assertEqual(a1.x, 1)
      self.assertEqual(a1.y, 2)

      # Override attributes take precedence over member attribute.
      with a1.override(y=3):
        self.assertEqual(a1.y, 3)
        with a1.override(y=4):
          self.assertEqual(a1.y, 4)
        self.assertEqual(a1.y, 3)
      self.assertEqual(a1.y, 2)

      # `contextual_override` takes precedence over contextual default.
      self.assertEqual(a1.z, 3)

      # Test nested contextual override with override_attrs=True (default).
      with contextual_override(
          y=4, z=4, override_attrs=True) as nested_override:
        self.assertEqual(
            nested_override,
            dict(
                x=ContextualOverride(3, cascade=False, override_attrs=False),
                y=ContextualOverride(4, cascade=False, override_attrs=True),
                z=ContextualOverride(4, cascade=False, override_attrs=True),
            ),
        )

        # Member attribute is not overriden as current scope does not override
        # `x``.
        self.assertEqual(a1.x, 1)

        # Member attribute is overriden.
        self.assertEqual(a1.y, 4)

        # `ContextualObject.override` takes precedence over
        # `contextual_override(override_attrs=True)`.
        with a1.override(y=3):
          self.assertEqual(a1.y, 3)
        self.assertEqual(a1.y, 4)

        # Member default is overriden.
        self.assertEqual(a1.z, 4)

      self.assertEqual(a1.y, 2)
      self.assertEqual(a1.z, 3)

    self.assertEqual(a1.y, 2)
    self.assertEqual(a1.z, -1)

  def test_context_cascade(self):
    class A(ContextualObject):
      x: int
      y: int = contextual_attribute()
      z: int = contextual_attribute(default=-1)

    a = A(1, 2)
    self.assertEqual(a.x, 1)
    self.assertEqual(a.y, 2)
    self.assertEqual(a.z, -1)

    with contextual_override(x=3, y=3, z=3, cascade=True):
      self.assertEqual(a.x, 1)
      self.assertEqual(a.y, 2)
      self.assertEqual(a.z, 3)

      # Outter `pg.contextual_override` takes precedence
      # over inner `pg.contextual_override` when cascade=True.
      with contextual_override(y=4, z=4, cascade=True):
        self.assertEqual(a.x, 1)
        self.assertEqual(a.y, 2)
        self.assertEqual(a.z, 3)

      with contextual_override(y=4, z=4, override_attrs=True):
        self.assertEqual(a.x, 1)
        self.assertEqual(a.y, 2)
        self.assertEqual(a.z, 3)

    self.assertEqual(a.x, 1)
    self.assertEqual(a.y, 2)
    self.assertEqual(a.z, -1)

    with contextual_override(x=3, y=3, z=3, cascade=True, override_attrs=True):
      self.assertEqual(a.x, 3)
      self.assertEqual(a.y, 3)
      self.assertEqual(a.z, 3)

      with contextual_override(y=4, z=4, override_attrs=True):
        self.assertEqual(a.x, 3)
        self.assertEqual(a.y, 3)
        self.assertEqual(a.z, 3)

    self.assertEqual(a.x, 1)
    self.assertEqual(a.y, 2)
    self.assertEqual(a.z, -1)

  def test_sym_inferred(self):
    class A(ContextualObject):
      x: int = 1
      y: int = contextual_attribute()

    a = A()
    with self.assertRaisesRegex(
        AttributeError, '.* is not found under its context'):
      _ = a.sym_inferred('y')
    self.assertIsNone(a.sym_inferred('y', default=None))

    with self.assertRaises(AttributeError):
      _ = a.sym_inferred('z')
    self.assertIsNone(a.sym_inferred('z', default=None))

  def test_weak_ref(self):
    class A(ContextualObject):
      x: int = 1

    a = A()
    self.assertIsNotNone(weakref.ref(a))


class ContextualAttributeTest(unittest.TestCase):
  """Tests for Component."""

  def test_contextualibute_access(self):

    class A(ContextualObject):
      x: int
      y: int = contextual_attribute()

    # Not okay: `A.x` is required.
    with self.assertRaisesRegex(TypeError, 'missing 1 required argument'):
      _ = A()

    # Okay: `A.y` is contextual.
    a = A(1)

    # `a.y` is not yet available from the context.
    with self.assertRaises(AttributeError):
      _ = a.y

    class B(ContextualObject):
      # Attributes with annotation will be treated as symbolic fields.
      p: int
      q: A = A(2)
      z: int = contextual_attribute()

    class C(ContextualObject):
      a: int
      b: B = B(2)
      y: int = 1
      z = 2

    c = C(1)
    b = c.b
    a = b.q

    # Test symbolic attributes declared from C.
    self.assertTrue(c.sym_hasattr('a'))
    self.assertTrue(c.sym_hasattr('b'))
    self.assertTrue(c.sym_hasattr('y'))
    self.assertFalse(c.sym_hasattr('z'))

    # Contextual access to c.y from a.
    self.assertEqual(a.y, 1)
    self.assertEqual(b.z, 2)

    # 'y' is not defined as an attribute in 'B'.
    with self.assertRaises(AttributeError):
      _ = b.y

    c.rebind(y=2)
    self.assertEqual(c.y, 2)
    self.assertEqual(a.y, 2)

    c.z = 3
    self.assertEqual(c.z, 3)
    self.assertEqual(b.z, 3)

  def test_to_html(self):
    class A(ContextualObject):
      x: int = 1
      y: int = contextual_attribute()

    def assert_content(html, expected):
      expected = inspect.cleandoc(expected).strip()
      actual = html.content.strip()
      if actual != expected:
        print(actual)
      self.assertEqual(actual.strip(), expected)

    self.assertIn(
        inspect.cleandoc(
            """
            .contextual-attribute {
              color: purple;
            }
            .unavailable-contextual {
              color: gray;
              font-style: italic;
            }
            """
        ),
        A().to_html().style_section,
    )

    assert_content(
        A().to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove a"><summary><div class="summary-title">A(...)</div></summary><div class="complex-value a"><details open class="pyglove int"><summary><div class="summary-name">x<span class="tooltip">x</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove contextual-attribute"><summary><div class="summary-name">y<span class="tooltip">y</span></div><div class="summary-title">ContextualAttribute(...)</div></summary><div class="unavailable-contextual">(not available)</div></details></div></details>
        """
    )

    class B(ContextualObject):
      z: Any
      y: int = 2

    b = B(A())
    assert_content(
        b.z.to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove a"><summary><div class="summary-title">A(...)</div></summary><div class="complex-value a"><details open class="pyglove int"><summary><div class="summary-name">x<span class="tooltip">x</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove contextual-attribute"><summary><div class="summary-name">y<span class="tooltip">y</span></div><div class="summary-title">ContextualAttribute(...)</div></summary><span class="simple-value int">2</span></details></div></details>
        """
    )


if __name__ == '__main__':
  unittest.main()
